// Display control functionality + support infrastructure
//This is an RTOS task.
//written by Webdust21 (2024-06-06)

#include <Arduino.h>
#include "display.h"
#include <ESP_Panel.h>
#include <ESP_Panel.cpp>        //@#&!@) need to #include this or somehow PIO can't find it??
#include <lvgl.h>
#include "lvgl_port_v9.h"
#include "../screens/screens.h"
HomeScreen *currentScreen;

uint8_t currentScreenId = 255;			// invalid entry just so the display init function runs to create a screen instance.
uint8_t ScreenId = screen_Main;			// start with "main" screen on boot.
lv_obj_t *mainContainer;

void rtos_hid_tft(void *pvParameters) {
	// FreeRTOS task function.  Must not "return" unless the thread is destroyed.
retry:
	//2024-06-18: Webdust21: noticed that the ESP_IO_Expander does not appear to be touched.  It's not configured in the
	//ESP_Panel_Conf setup, and the corresponding pin definitions are disabled (== -1).
	//This port expander gives access to some very valuable functions, namely, being able to force-reset the GT911
	//touch controller and TFT controller to ensure they are ready upon startup.  Without this, it sometimes refuses
	//to function upon power-up.

    // Set up the TFT LCD
	log_i("Initialize ESP_Panel Library");
    ESP_Panel *panel = new ESP_Panel();
    if (!panel->init()) {log_e("failed initializing ESP_Panel.");}
#if LVGL_PORT_AVOID_TEAR
    // When avoid tearing function is enabled, configure the RGB bus according to the LVGL configuration
    ESP_PanelBus_RGB *rgb_bus = static_cast<ESP_PanelBus_RGB *>(panel->getLcd()->getBus());
    rgb_bus->configRgbFrameBufferNumber(LVGL_PORT_DISP_BUFFER_NUM);
    rgb_bus->configRgbBounceBufferSize(LVGL_PORT_RGB_BOUNCE_BUFFER_SIZE);
#endif
	log_i("Start ESP_Panel Library");
    if (!panel->begin()) {
		log_e("failed starting ESP_Panel.  Self-restart.");
		panel->del();	// otherwise it won't re-init the I2C bus.
		delay(1000);
		goto retry;
	}

    log_i("Initialize LVGL");
    lvgl_port_init(panel->getLcd(), panel->getTouch());

	// Set the "container" for all pages.  NOTE: This can easily be set to a window object to
	//"contain" the pages in a tabbed UI if desired.
	mainContainer = lv_scr_act();

	TickType_t xLastWakeTime = xTaskGetTickCount();
	uint32_t task_delay_ms = LVGL_PORT_TASK_MAX_DELAY_MS;
	unsigned long uiPollTimer = millis() + 100;
	while (true) {
		//NOTE: the "lvgl_port_v8" puts LVGL tasking in a separate RTOS thread.  This means that we do
		//not have to call the LVGL Timer functionality to keep the screen updated.  
		//We just have to keep track of our responsibilities here.
		// -> 2024-07-12: absolutely no reason to have that split up.  Merged task timer functionality
		//with this loop here.
        if (lvgl_port_lock(-1)) {
            task_delay_ms = lv_timer_handler();

			// Every 100mS (or so!) call UI update function...
			if (millis() > uiPollTimer) {
				uiPollTimer = millis() + 100;

				// Call the "update" function on the currently active screen
				//WILL CRASH if there is no valid screen loaded.
				// Does the screen class need changed?
				if (currentScreenId != ScreenId) {
					// yes, the screen needs to be changed.
					//Trigger the deconstructor in the screen class object.  THIS calls "lv_obj_clean" on the container
					delete currentScreen;
					
					// Create a new screen class with the requested screen.
					switch (currentScreenId = ScreenId)	{	// assign & check
						case screen_Main:	currentScreen = new HomeScreen_Main(mainContainer); break;
						//Add more screens here as necessary
					}
				} else {
					// call the Update function on the screen class to handle the necessary functionality.
					currentScreen->Update();
				}
			}

            lvgl_port_unlock();

			if (task_delay_ms > LVGL_PORT_TASK_MAX_DELAY_MS) {
				task_delay_ms = LVGL_PORT_TASK_MAX_DELAY_MS;
			} else if (task_delay_ms < LVGL_PORT_TASK_MIN_DELAY_MS) {
				task_delay_ms = LVGL_PORT_TASK_MIN_DELAY_MS;
			}
        }
		xTaskDelayUntil(&xLastWakeTime, pdMS_TO_TICKS(task_delay_ms));  // vTaskDelay(pdMS_TO_TICKS(task_delay_ms));
	}

	// if the (while(true)) above is ever changed to allow an exit of the loop,
	//we must delete "our" thread to avoid a system crash.
	vTaskDelete(nullptr);
}
